import math
from collections import namedtuple


def findattr(obj, name):
    """Similar to builtin getattr(obj, name) but more forgiving to
    whitespace and capitalization.
    
    """
    # Come up with several options
    py_name = name.replace('-', '_').replace(' ', '_').replace("'", "")
    camel_case = "".join([s.capitalize() for s in py_name.split('_')])
    if hasattr(obj, py_name):
        # Direct lookup
        attr = getattr(obj, py_name)
    elif hasattr(obj, camel_case):
        # CamelCase lookup
        attr = getattr(obj, camel_case)
    else:
        raise AttributeError(f'{obj} has no attribute {name}')
    return attr

def mod_str(modifier):
    """Converts a modifier to a string, eg 2 -> '+2'."""
    if modifier > 0:
        mod_str = '+' + str(modifier)
    else:
        mod_str = str(modifier)
    return mod_str


AbilityScore = namedtuple('AbilityScore', ('value', 'modifier', 'saving_throw'))


class Ability():
    ability_name = None
    
    def __init__(self, default_value=10):
        self.default_value = default_value
    
    def __set_name__(self, character, name):
        self.ability_name = name
    
    def _check_dict(self, obj):
        if not hasattr(obj, '_ability_scores'):
            # No ability score dictionary exists
            obj._ability_scores = {
                self.ability_name: self.default_value
            }
        elif self.ability_name not in obj._ability_scores.keys():
            # ability score dictionary exists but doesn't have this ability
            obj._ability_scores[self.ability_name] = self.default_value
    
    def __get__(self, character, Character):
        self._check_dict(character)
        score = character._ability_scores[self.ability_name]
        modifier = math.floor((score - 10) / 2)
        # Check for proficiency
        saving_throw = modifier
        if self.ability_name is not None:
            is_proficient = (self.ability_name in character.saving_throw_proficiencies)
            if is_proficient:
                saving_throw += character.proficiency_bonus
        # Create the named tuple
        value = AbilityScore(modifier=modifier, value=score, saving_throw=saving_throw)
        return value
    
    def __set__(self, character, val):
        self._check_dict(character)
        character._ability_scores[self.ability_name] = val
        self.value = val


class Skill():
    """An ability-based skill, such as athletics."""
    
    def __init__(self, ability):
        self.ability_name = ability
    
    def __set_name__(self, character, name):
        self.skill_name = name
        self.character = character
    
    def __get__(self, character, owner):
        ability = getattr(character, self.ability_name)
        modifier = ability.modifier
        # Check for proficiency
        is_proficient = self.skill_name in character.skill_proficiencies
        if is_proficient:
            modifier += character.proficiency_bonus
        return modifier



class Background():
    name = "Generic background"
    skill_proficiencies = ()
    languages = ()


class Acolyte(Background):
    name = "Acolyte"
    skill_proficiencies = ('insight', 'religion')
    languages = ("[choose one]", "[choose one]")


class Charlatan(Background):
    name = "Charlatan"
    skill_proficiencies = ('deception', 'sleight of hand')


class Criminal(Background):
    name = "Criminal"
    skill_proficiencies = ('deception', 'stealth')


class Spy(Criminal):
    name = "Spy"


class Entertainer(Background):
    name = "Entertainer"
    skill_proficiencies = ('acrobatic', 'performance')


class Gladiator(Entertainer):
    name = "Gladiator"


class FolkHero(Background):
    name = "Folk Hero"
    skill_proficiencies = ('animal handling', 'survival')


class GuildArtisan(Background):
    name = "Guild Artisan"
    skill_proficiencies = ('insight', 'persuasion')
    languages = ("[choose one]", "[choose one]")


class GuildMerchant(GuildArtisan):
    name = "Guild Merchant"


class Hermit(Background):
    name = "Hermit"
    skill_proficiencies = ("medicine", "religion")
    languages = ("[choose one]")


class Noble(Background):
    name = "Noble"
    skill_proficiencies = ("history", 'persuasion')
    languages = ("[choose one]", )


class Knight(Noble):
    name = "Knight"


class Outlander(Background):
    name = "Outlander"
    skill_proficiencies = ('athletics', 'survival')
    languages = ("[choose one]", )


class Sage(Background):
    name = "Sage"
    skill_proficiencies = ('arcana', 'history')
    languages = ("[choose one]", '[choose one]')


class Sailor(Background):
    name = "Sailor"
    skill_proficiencies = ('athletics', 'perception')


class Pirate(Sailor):
    name = "Pirate"


class Soldier(Background):
    name = "Soldier"
    skill_proficiencies = ('athletics', 'intimidation')


class Urchin(Background):
    name = "Urchin"
    skill_proficiencies = ('sleight of hand', 'stealth')



"""Tools for describing a player character."""

import re
import warnings

from .stats import Ability, Skill, findattr
from .dice import read_dice_str
from . import weapons, race, spells, armor
from .weapons import Weapon
from .armor import Armor, NoArmor, Shield, NoShield

dice_re = re.compile('(\d+)d(\d+)')

__all__ = ('Barbarian', 'Bard', 'Cleric', 'Druid', 'Fighter', 'Monk',
           'Paladin', 'Ranger', 'Rogue', 'Sorceror', 'Warlock', 'Wizard', )

class Character():
    """A generic player character. Intended to be subclasses by the
    various classes.
    
    """
    # General attirubtes
    name = ""
    class_name = ""
    player_name = ""
    background = ""
    level = 1
    alignment = "Neutral"
    race = None
    xp = 0
    # Hit points
    hp_max = 10
    hit_dice_faces = 2
    # Base stats (ability scores)
    strength = Ability()
    dexterity = Ability()
    constitution = Ability()
    intelligence = Ability()
    wisdom = Ability()
    charisma = Ability()
    saving_throw_proficiencies = []
    skill_proficiencies = tuple()
    class_skill_choices = tuple()
    num_skill_choices = 2
    weapon_proficiencies = tuple()
    proficiencies_extra = tuple()
    languages = ""
    # Skills
    acrobatics = Skill(ability='dexterity')
    animal_handling = Skill(ability='wisdom')
    arcana = Skill(ability='intelligence')
    athletics = Skill(ability='strength')
    deception = Skill(ability='charisma')
    history = Skill(ability='intelligence')
    insight = Skill(ability='wisdom')
    intimidation = Skill(ability='charisma')
    investigation = Skill(ability='intelligence')
    medicine = Skill(ability='wisdom')
    nature = Skill(ability='intelligence')
    perception = Skill(ability='wisdom')
    performance = Skill(ability='charisma')
    persuasian = Skill(ability='charisma')
    religion = Skill(ability='intelligence')
    sleight_of_hand = Skill(ability='dexterity')
    stealth = Skill(ability='dexterity')
    survival = Skill(ability='wisdom')
    # Characteristics
    attacks_and_spellcasting = ""
    personality_traits = ""
    ideals = ""
    bonds = ""
    flaws = ""
    features_and_traits = ""
    # Inventory
    cp = 0
    sp = 0
    ep = 0
    gp = 0
    pp = 0
    equipment = ""
    weapons = [] # Replaced in __init__ constructor
    armor = None
    shield = None
    _proficiencies_text = tuple()
    # Magic
    spellcasting_ability = None
    spells = tuple()
    spells_prepared = tuple()
    
    def __init__(self, **attrs):
        """Takes a bunch of attrs and passes them to ``set_attrs``"""
        self.weapons = []
        self.set_attrs(**attrs)
    
    def __str__(self):
        return self.name
    
    def __repr__(self):
        return f"<{self.class_name}: {self.name}>"
    
    @property
    def speed(self):
        return self.race.speed
    
    def set_attrs(self, **attrs):
        """Bulk setting of attributes. Useful for loading a character from a
        dictionary."""
        for attr, val in attrs.items():
            if attr == 'weapons':
                # Treat weapons specially
                for weap in val:
                    self.wield_weapon(weap)
            elif attr == 'race':
                MyRace = findattr(race, val)
                self.race = MyRace()
            elif attr == 'armor':
                self.wear_armor(val)
            elif attr == 'shield':
                self.wield_shield(val)
            elif (attr == 'spells') or (attr == 'spells_prepared'):
                # Create a list of actual spell objects
                _spells = []
                for spell_name in val:
                    try:
                        _spells.append(findattr(spells, spell_name))
                    except AttributeError:
                        msg = f'Spell "{spell_name}" not defined. Please add it to ``spells.py``'
                        warnings.warn(msg)
                        # Create temporary spell
                        _spells.append(spells.create_spell(name=spell_name, level=9))
                        # raise AttributeError(msg)
                # Sort by name
                _spells.sort(key=lambda spell: spell.name)
                # Save list of spells to character atribute
                if attr == 'spells':
                    # Instantiate them all for the spells list
                    self.spells = tuple(S() for S in _spells)
                else:
                    self.spells_prepared = tuple(_spells)
            else:
                if not hasattr(self, attr):
                    warnings.warn(f"Setting unknown character attribute {attr}",
                                  RuntimeWarning)
                # Lookup general attributes
                setattr(self, attr, val)
    
    @property
    def is_spellcaster(self):
        result = (self.spellcasting_ability is not None)
        return result
    
    @property
    def spell_save_dc(self):
        ability_mod = getattr(self, self.spellcasting_ability).modifier
        return (8 + self.proficiency_bonus + ability_mod)
    
    @property
    def spell_attack_bonus(self):
        ability_mod = getattr(self, self.spellcasting_ability).modifier
        return (self.proficiency_bonus + ability_mod)
    
    def spell_slots(self, spell_level):
        """How many spells slots are available for this spell level."""
        return self.spell_slots_by_level[self.level][spell_level]
    
    def is_proficient(self, weapon: Weapon):
        """Is the character proficient with this item?
        
        Considers class proficiencies and race proficiencies.
        
        Parameters
        ----------
        weapon
          The weapon to be tested for proficiency.
        
        """
        all_proficiencies = tuple(self.weapon_proficiencies)
        all_proficiencies += tuple(getattr(self.race, 'weapon_proficiencies', tuple()))
        is_proficient = any((isinstance(weapon, W) for W in all_proficiencies))
        return is_proficient
    
    @property
    def proficiencies_text(self):
        final_text = ""
        all_proficiencies = self._proficiencies_text
        if self.race is not None:
            all_proficiencies += self.race.proficiencies_text
        all_proficiencies += self.proficiencies_extra
        # Create a single string out of all the proficiencies
        for txt in all_proficiencies:
            if not final_text:
                # Capitalize the first entry
                txt = txt.capitalize()
            else:
                # Put a comma first
                txt = ", " + txt
                # Add this item to the list text
            final_text += txt
        # Add a period at the end
        final_text += '.'
        return final_text
    
    def wear_armor(self, new_armor):
        """Accepts a string or Armor class and replaces the current armor.
        
        If a string is given, then a subclass of
        :py:class:`~dungeonsheets.armor.Armor` is retrived from the
        ``armor.py`` file. Otherwise, an subclass of
        :py:class:`~dungeonsheets.armor.Armor` can be provided
        directly.
        
        """
        if new_armor not in ('', None):
            try:
                NewArmor = findattr(armor, new_armor)
            except AttributeError:
                # Not a string, so just treat it as Armor
                NewArmor = new_armor
            self.armor = NewArmor()
    
    def wield_shield(self, shield):
        """Accepts a string or Shield class and replaces the current armor.
        
        If a string is given, then a subclass of
        :py:class:`~dungeonsheets.armor.Shield` is retrived from the
        ``armor.py`` file. Otherwise, an subclass of
        :py:class:`~dungeonsheets.armor.Shield` can be provided
        directly.
        
        """
        if shield not in ('', None):
            try:
                NewShield = findattr(armor, shield)
            except AttributeError:
                # Not a string, so just treat it as Armor
                NewShield = shield
            self.shield = NewShield()
    
    def wield_weapon(self, weapon):
        """Accepts a string and adds it to the list of wielded weapons.
        
        Parameters
        ----------
        weapon : str
          Case-insensitive string with a name of the weapon.
        
        """
        # Retrieve the weapon class from the weapons module
        try:
            NewWeapon = findattr(weapons, weapon)
        except AttributeError:
            raise AttributeError(f'Weapon {weapon} is not defined')
        weapon_ = NewWeapon()
        # Set weapon attributes based on character
        if weapon_.is_finesse:
            ability_mod = max(self.strength.modifier, self.dexterity.modifier)
        else:
            ability_mod = getattr(self, weapon_.ability).modifier
        weapon_.attack_bonus += ability_mod
        weapon_.bonus_damage += ability_mod
        # Check for prifiency
        if self.is_proficient(weapon_):
            weapon_.attack_bonus += self.proficiency_bonus
        # Save it to the array
        self.weapons.append(weapon_)
    
    @property
    def hit_dice(self):
        """What type and how many dice to use for re-gaining hit points.
        
        To change, set hit_dice_num and hit_dice_faces."""
        return f"{self.level}d{self.hit_dice_faces}"
    
    @property
    def proficiency_bonus(self):
        if self.level < 5:
            prof = 2
        elif 5 <= self.level < 9:
            prof = 3
        elif 9 <= self.level < 13:
            prof = 4
        elif 13 <= self.level < 17:
            prof = 5
        elif 17 <= self.level:
            prof = 6
        return prof
    
    @property
    def armor_class(self):
        """Armor class, including contributions from worn armor and shield."""
        # Retrieve current armor (or a generic armor substitute)
        armor = self.armor if self.armor is not None else NoArmor()
        shield = self.shield if self.shield is not None else NoShield()
        # Calculate and apply modifiers
        if armor.dexterity_mod_max is None:
            modifier = self.dexterity.modifier
        else:
            modifier = min(self.dexterity.modifier, armor.dexterity_mod_max)
        # Calculate final armor class
        ac = armor.base_armor_class + shield.base_armor_class + modifier
        return ac


class Barbarian(Character):
    class_name = 'Barbarian'
    hit_dice_faces = 12
    saving_throw_proficiencies = ('strength', 'constitution')
    _proficiencies_text = ('light armor', 'medium armor', 'shields',
                           'simple weapons', 'martial weapons')
    weapon_proficiencies = (weapons.simple_weapons + weapons.martial_weapons)
    class_skill_choices = ('Animal Handling', 'Athletics',
                           'Intimidation', 'Nature', 'Perception', 'Survival')


class Bard(Character):
    class_name = 'Bard'
    hit_dice_faces = 8
    saving_throw_proficiencies = ('dexterity', 'charisma')
    _proficiencies_text = (
        'Light armor', 'simple weapons', 'hand crossbows', 'longswords',
        'rapiers', 'shortswords', 'three musical instruments of your choice')
    weapon_proficiencies = ((weapons.HandCrossbow, weapons.Longsword,
                            weapons.Rapier, weapons.Shortsword) +
                           weapons.simple_weapons)
    class_skill_choices = ('Acrobatics', 'Animal Handling', 'Arcana',
                           'Athletics', 'Deception', 'History', 'Insight',
                           'Intimidation', 'Investigation', 'Medicine', 'Nature',
                           'Perception', 'Performance', 'Persuasion', 'Religion',
                           'Sleight of Hand', 'Stealth', 'Survival')
    num_skill_choices = 3


class Cleric(Character):
    class_name = 'Cleric'
    hit_dice_faces = 8
    saving_throw_proficiencies = ('wisdom', 'charisma')
    _proficiencies_text = ('light armor', 'medium armor', 'shields',
                          'all simple weapons')
    weapon_proficiencies = weapons.simple_weapons
    class_skill_choices = ('History', 'Insight', 'Medicine',
                           'Persuasion', 'Religion')


class Druid(Character):
    class_name = 'Druid'
    hit_dice_faces = 8
    saving_throw_proficiencies = ('intelligence', 'wisdom')
    _proficiencies_text = (
        'Light armor', 'medium armor',
        'shields (druids will not wear armor or use shields made of metal)',
        'clubs', 'daggers', 'darts', 'javelins', 'maces', 'quarterstaffs',
        'scimitars', 'sickles', 'slings', 'spears')
    weapon_proficiencies = (weapons.Club, weapons.Dagger, weapons.Dart,
                           weapons.Javelin, weapons.Mace, weapons.Quarterstaff,
                           weapons.Scimitar, weapons.Sickle, weapons.Sling, weapons.Spear)
    class_skill_choices = ('Arcana', 'Animal Handling', 'Insight',
                           'Medicine', 'Nature', 'Perception', 'Religion', 'Survival')


class Fighter(Character):
    class_name = 'Fighter'
    hit_dice_faces = 10
    saving_throw_proficiencies = ('strength', 'constitution')
    _proficiencies_text = ('All armar', 'shields', 'simple weapons', 'martial weapons')
    weapon_proficiencies = weapons.simple_weapons + weapons.martial_weapons
    class_skill_choices = ('Acrobatics', 'Animal Handling',
                           'Athletics', 'History', 'Insight', 'Intimidation', 'Perception',
                           'Survival')


class Monk(Character):
    class_name = 'Monk'
    hit_dice_faces = 8
    saving_throw_proficiencies = ('strength', 'dexterity')
    _proficiencies_text = (
        'simple weapons', 'shortswords',
        "one type of artisan's tools or one musical instrument")
    weapon_proficiencies = (weapons.Shortsword,) + weapons.simple_weapons
    class_skill_choices = ('Acrobatics', 'Athletics', 'History', 'Insight', 'Religion', 'Stealth')

class Paladin(Character):
    class_name = 'Paladin'
    hit_dice_faces = 10
    saving_throw_proficiencies = ('wisdom', 'charisma')
    _proficiencies_text = ('All armor', 'shields', 'simple weapons',
                          'martial weapons')
    weapon_proficiencies = weapons.simple_weapons + weapons.martial_weapons
    class_skill_choices = ("Athletics", 'Insight', 'Intimidation',
                           'Medicine', 'Persuasion', 'Religion')


class Ranger(Character):
    class_name = 'Ranger'
    hit_dice_faces = 10
    saving_throw_proficiencies = ('strength', 'dexterity')
    _proficiencies_text = ("light armor", "medium armor", "shields",
                           "simple weapons", "martial weapons")
    weapon_proficiencies = weapons.simple_weapons + weapons.martial_weapons
    class_skill_choices = ('Animal Handling', 'Athletics', 'Insight',
                           'Investigation', 'Nature', 'Perception', 'Stealth', 'Survival')
    num_skill_choices = 3


class Rogue(Character):
    class_name = 'Rogue'
    hit_dice_faces = 8
    saving_throw_proficiencies = ('dexterity', 'intelligence')
    _proficiencies_text = (
        'light armor', 'simple weapons', 'hand crossbows', 'longswords',
        'rapiers', 'shortswords', "thieves' tools")
    weapon_proficiencies = (weapons.HandCrossbow, weapons.Longsword,
                           weapons.Rapier, weapons.Shortsword) + weapons.simple_weapons
    class_skill_choices = ('Acrobatics', 'Athletics', 'Deception',
                           'Insight', 'Intimidation', 'Investigation', 'Perception',
                           'Performance', 'Persuasion', 'Sleight of Hand', 'Stealth')


class Sorceror(Character):
    class_name = 'Sorceror'
    hit_dice_faces = 6
    saving_throw_proficiencies = ('constitution', 'charisma')
    _proficiencies_text = ('daggers', 'darts', 'slings',
                           'quarterstaffs', 'light crossbows')
    weapon_proficiencies = (weapons.Dagger, weapons.Dart,
                           weapons.Sling, weapons.Quarterstaff,
                           weapons.LightCrossbow)
    class_skill_choices = ('Arcana', 'Deception', 'Insight',
                           'Intimidation' ,'Persuasion', 'Religion')


class Warlock(Character):
    class_name = 'Warlock'
    hit_dice_faces = 8
    saving_throw_proficiencies = ('wisdom', 'charisma')
    _proficiencies_text = ("light Armor", "simple weapons")
    class_skill_choices = ('Arcana', 'Deception', 'History',
                           'Intimidation', 'Investigation', 'Nature', 'Religion')
    weapon_proficiencies = weapons.simple_weapons
    spellcasting_ability = 'charisma'
    spell_slots_by_level = {
        1:  (2, 1, 0, 0, 0, 0, 0, 0, 0, 0),
        2:  (2, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        3:  (2, 0, 2, 0, 0, 0, 0, 0, 0, 0),
        4:  (3, 0, 2, 0, 0, 0, 0, 0, 0, 0),
        5:  (3, 0, 0, 3, 0, 0, 0, 0, 0, 0),
        6:  (3, 0, 0, 3, 0, 0, 0, 0, 0, 0),
        7:  (3, 0, 0, 0, 2, 0, 0, 0, 0, 0),
        8:  (3, 0, 0, 0, 2, 0, 0, 0, 0, 0),
        9:  (3, 0, 0, 0, 0, 2, 0, 0, 0, 0),
        10: (4, 0, 0, 0, 0, 2, 0, 0, 0, 0),
        11: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        12: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        13: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        14: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        15: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        16: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        17: (4, 0, 0, 0, 0, 4, 0, 0, 0, 0),
        18: (4, 0, 0, 0, 0, 4, 0, 0, 0, 0),
        19: (4, 0, 0, 0, 0, 4, 0, 0, 0, 0),
        20: (4, 0, 0, 0, 0, 4, 0, 0, 0, 0),
    }


class Wizard(Character):
    class_name = 'Wizard'
    hit_dice_faces = 6
    saving_throw_proficiencies = ('intelligence', 'wisdom')
    _proficiencies_text = ('daggers', 'darts', 'slings',
                          'quarterstaffs', 'light crossbows')
    weapon_proficiencies = (weapons.Dagger, weapons.Dart,
                           weapons.Sling, weapons.Quarterstaff,
                           weapons.LightCrossbow)
    class_skill_choices = ('Arcana', 'History', 'Investigation',
                           'Medicine', 'Religion')
    spellcasting_ability = 'intelligence'
    spell_slots_by_level = {
        # char_lvl: (cantrips, 1st, 2nd, 3rd, ...)
        1:  (3, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        2:  (3, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        3:  (3, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        4:  (4, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        5:  (4, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        6:  (4, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        7:  (4, 4, 3, 3, 1, 0, 0, 0, 0, 0),
        8:  (4, 4, 3, 3, 2, 0, 0, 0, 0, 0),
        9:  (4, 4, 3, 3, 3, 1, 0, 0, 0, 0),
        10: (5, 4, 3, 3, 3, 2, 0, 0, 0, 0),
        11: (5, 4, 3, 3, 3, 2, 1, 0, 0, 0),
        12: (5, 4, 3, 3, 3, 2, 1, 0, 0, 0),
        13: (5, 4, 3, 3, 3, 2, 1, 1, 0, 0),
        14: (5, 4, 3, 3, 3, 2, 1, 1, 0, 0),
        15: (5, 4, 3, 3, 3, 2, 1, 1, 1, 0),
        16: (5, 4, 3, 3, 3, 2, 1, 1, 1, 0),
        17: (5, 4, 3, 3, 3, 2, 1, 1, 1, 1),
        18: (5, 4, 3, 3, 3, 3, 1, 1, 1, 1),
        19: (5, 4, 3, 3, 3, 3, 2, 1, 1, 1),
        20: (5, 4, 3, 3, 3, 3, 2, 2, 1, 1),
    }



class Shield():
    """A shield that can be worn on one hand."""
    name = "Shield"
    cost = "10 gp"
    base_armor_class = 2
    
    def __str__(self):
        return self.name


class NoShield(Shield):
    """If a character is carrying no shield."""
    name = "No shield"
    cost = "0"
    base_armor_class = 0
    
    def __str__(self):
        return self.name


class Armor():
    """A piece of armor that can be worn.
    
    Attributes
    ----------
    name : str
      Human-readable name for this armor.
    cost : str
      Cost and currency for this armor.
    base_armor_class : int
      Armor class granted before modifiers.
    dexterity_mod_max : int
      How much dexterity can the user contribute. ``0`` for no
      dexterity modifier, ``None`` for unlimited dexterity modifier.
    strength_required : int
      Minimum strength needed to use this armor properly.
    stealth_disadvantage : bool
      If true, the armor causes disadvantage on stealth rolls.
    weight : int
      In lbs.
    
    """
    name = "Unknown Armor"
    cost = "0 gp"
    base_armor_class = 10
    dexterity_mod_max = None
    strength_required = None
    stealth_disadvantage = False
    weight = 0 # In lbs
    
    def __str__(self):
        return self.name


class NoArmor(Armor):
    name = "No armor"


class LightPaddedArmor(Armor):
    name = "Light padded armor"
    cost = "5 gp"
    base_armor_class = 11
    weight = 8
    stealth_disadvantage = True


class LightLeatherArmor(Armor):
    name = "Light leather armor"
    cost = "10 gp"
    base_armor_class = 11
    weight = 10


class LightStuddedArmor(Armor):
    name = "Light studded armor"
    cost = "45 gp"
    base_armor_class = 12
    weight = 13


class MediumHideArmor(Armor):
    name = "Medium hide armor"
    cost = "10 gp"
    base_armor_class = 12
    dexterity_mod_max = 2
    weight = 12


class MediumChainShirtArmor(Armor):
    name = "Medium chain shirt armor"
    cost = "50 gp"
    base_armor_class = 13
    dexterity_mod_max = 2
    weight = 20


class MediumScaleMailArmor(Armor):
    name = "Medium scale mail armor"
    cost = "50 gp"
    base_armor_class = 14
    dexterity_mod_max = 2
    stealth_disadvantage = True
    weight = 45


class MediumBrassplateArmor(Armor):
    name = "Medium brassplate armor"
    cost = "400 gp"
    base_armor_class = 14
    dexterity_mod_max = 2
    weight = 20


class MediumHalfPlateArmor(Armor):
    name = "Medium half plate armor"
    cost = "750 gp"
    base_armor_class = 15
    dexterity_mod_max = 2
    stealth_disadvantage = True
    weight = 40


class HeavyRingMailArmor(Armor):
    name = "Heavy ring mail armor"
    cost = "30 gp"
    base_armor_class = 14
    dexterity_mod_max = 0
    stealth_disadvantage = True
    weight = 40


class HeavyChainMailArmor(Armor):
    name = "Heavy chain mail armor"
    cost = "75 gp"
    base_armor_class = 16
    dexterity_mod_max = 0
    strength_required = 13
    stealth_disadvantage = True
    weight = 55


class HeavySplintArmor(Armor):
    name = "Heavy splint armor"
    cost = "200 gp"
    base_armor_class = 17
    dexterity_mod_max = 0
    strength_required = 15
    stealth_disadvantage = True
    weight = 60


class HeavyPlateArmor(Armor):
    name = "Heavy splint armor"
    cost = "1,500 gp"
    base_armor_class = 18
    dexterity_mod_max = 0
    strength_required = 15
    stealth_disadvantage = True
    weight = 65



from .stats import mod_str

class Weapon():
    name = ""
    cost = "0 gp"
    base_damage = "1d4"
    bonus_damage = 0
    damage_type = "piercing"
    attack_bonus = 0
    weight = 1 # In lbs
    properties = "Light"
    ability = 'strength'
    is_finesse = False
    
    @property
    def damage(self):
        dam_str = str(self.base_damage)
        if self.bonus_damage != 0:
            dam_str += ' ' + mod_str(self.bonus_damage)
        return dam_str


class Club(Weapon):
    name = "Club"
    cost = "1 sp"
    base_damage = "1d4"
    damage_type = "bludgeoning"
    weight = 2
    properties = "Light"
    ability = 'strength'


class Dagger(Weapon):
    name = "Dagger"
    cost = "2 gp"
    base_damage = "1d4"
    damage_type = "piercing"
    weight = 1
    properties = "Finesse, light, thrown (range 20/60)"
    is_finesse = True
    ability = 'strength'


class Greatclub(Weapon):
    name = "Greatclub"
    cost = "2 sp"
    base_damage = "1d8"
    damage_type = "bludgeoning"
    weight = 10
    properties = "Two-handed"
    ability = 'strength'


class Handaxe(Weapon):
    name = "Handaxe"
    cost = "5 gp"
    base_damage = "1d6"
    damage_type = "slashing"
    weight = 2
    properties = "Light, thrown (range 20/60)"
    ability = 'strength'


class Javelin(Weapon):
    name = "Javelin"
    cost = "5 sp"
    base_damage = "1d6"
    damage_type = "piercing"
    weight = 2
    properties = "Thrown (range 30/120)"
    ability = 'strength'


class LightHammer(Weapon):
    name = "Light hammer"
    cost = "2 gp"
    base_damage = "1d4"
    damage_type = "bludgeoning"
    weight = 2
    properties = "Light, thrown (range 20/60)"
    ability = 'strength'


class Mace(Weapon):
    name = "Mace"
    cost = "5 gp"
    base_damage = "1d6"
    damage_type = "bludgeoning"
    weight = 4
    properties = ""
    ability = 'strength'


class Quarterstaff(Weapon):
    name = "Quarterstaff"
    cost = "2 sp"
    base_damage = "1d6"
    damage_type = "bludgeoning"
    weight = 4
    properties = "Versatile (1d8)"
    ability = 'strength'


class Sickle(Weapon):
    name = "Sickle"
    cost = "1 gp"
    base_damage = "1d4"
    damage_type = "slashing"
    weight = 2
    properties = "Light"
    ability = 'strength'


class Spear(Weapon):
    name = "Spear"
    cost = "1 gp"
    base_damage = "1d6"
    damage_type = "piercing"
    weight = 3
    properties = "Thrown (range 20/60), versatile (1d8)"
    ability = 'strength'


class LightCrossbow(Weapon):
    name = "Light crossbow"
    cost = "25 gp"
    base_damage = "1d8"
    damage_type = "piercing"
    weight = 5
    properties = "Ammunition (range 80/320, loading, two-handed"
    ability = 'dexterity'


class Dart(Weapon):
    name = "Dart"
    cost = "5 cp"
    base_damage = "1d4"
    damage_type = "piercing"
    weight = 0.25
    properties = "Finesse, thrown (range 20/60)"
    is_finesse = True
    ability = 'dexterity'


class Shortbow(Weapon):
    name = "Shortbow"
    cost = "25 gp"
    base_damage = "1d6"
    damage_type = "piercing"
    weight = 2
    properties = "Ammunition (range 80/320), two-handed"
    ability = 'dexterity'


class Sling(Weapon):
    name = "Sling"
    cost = "1 sp"
    base_damage = "1d4"
    damage_type = "bludgeoning"
    weight = 0
    properties = "Ammunition (range 30/120)"
    ability = 'dexterity'


class Battleaxe(Weapon):
    name = "Battleaxe"
    cost = "10 gp"
    base_damage = "1d8"
    damage_type = "slashing"
    weight = 4
    properties = "Versatile (1d10)"
    ability = 'strength'


class Flail(Weapon):
    name = "Flail"
    cost = "10gp"
    base_damage = "1d8"
    damage_type = "bludgeoning"
    weight = 2
    properties = ""
    ability = 'strength'


class Glaive(Weapon):
    name = "Glaive"
    cost = "20 gp"
    base_damage = "1d10"
    damage_type = "slashing"
    weight = 6
    properties = "Heavy, reach, two-handed"
    ability = 'strength'


class Greataxe(Weapon):
    name = "Greataxe"
    cost = "30 gp"
    base_damage = "1d12"
    damage_type = "slashing"
    weight = 7
    properties = "Heavy, two-handed"
    ability = 'strength'


class Greatsword(Weapon):
    name = "Greatsword"
    cost = "50 gp"
    base_damage = "2d6"
    damage_type = "slashing"
    weight = 6
    properties = "Heavy, two-handed"
    ability = 'strength'


class Halberd(Weapon):
    name = "Halberd"
    cost = "20 gp"
    base_damage = "1d10"
    damage_type = "slashing"
    weight = 6
    properties = "Heavy, reach, two-handed"
    ability = 'strength'


class Lance(Weapon):
    name = "Lance"
    cost = "10gp"
    base_damage = "1d12"
    damage_type = "piercing"
    weight = 6
    properties = "Reach, special"
    ability = 'strength'


class Longsword(Weapon):
    name = "Longsword"
    cost = "15 gp"
    base_damage = "1d8"
    damage_type = "slashing"
    weight = 3
    properties = "Versatile (1d10)"
    ability = 'strength'


class Maul(Weapon):
    name = "Maul"
    cost = "10 gp"
    base_damage = "2d6"
    damage_type = "bludgeoning"
    weight = 10
    properties = "Heavy, two-handed"
    ability = 'strength'


class Morningstar(Weapon):
    name = "Morningstar"
    cost = "15 gp"
    base_damage = "1d8"
    damage_type = "piercing"
    weight = 4
    properties = ""
    ability = 'strength'


class Pike(Weapon):
    name = "Pike"
    cost = "5 gp"
    base_damage = "1d10"
    damage_type = "piercing"
    weight = 18
    properties = "Heavy, reach, two-handed"
    ability = 'strength'


class Rapier(Weapon):
    name = "Rapier"
    cost = "25 gp"
    base_damage = "1d8"
    damage_type = "piercing"
    weight = 2
    properties = "Finesse"
    is_finesse = True
    ability = 'strength'


class Scimitar(Weapon):
    name = "Scimitar"
    cost = "25 gp"
    base_damage = "1d6"
    damage_type = "slashing"
    weight = 3
    properties = "Finesse, light"
    is_finesse = True
    ability = 'strength'


class Shortsword(Weapon):
    name = "Shortsword"
    cost = "10 gp"
    base_damage = "1d6"
    damage_type = "piercing"
    weight = 0
    properties = "Finesse, light"
    is_finesse = True
    ability = 'strength'


class ThrowingHammer(Weapon):
    name = "Throwing Hammer"
    cost = "15 gp"
    base_damage = '1d6'
    damage_type = "bludgeoning"
    weight = 4
    properties = "Thrown (range 60/120)"
    ability = "strength"


class Trident(Weapon):
    name = "Trident"
    cost = "5 gp"
    base_damage = "1d6"
    damage_type = "piercing"
    weight = 4
    properties = "Thrown (range 20/60), versatile (1d8)"
    ability = 'strength'


class WarPick(Weapon):
    name = "War pick"
    cost = "5 gp"
    base_damage = "1d8"
    damage_type = "piercing"
    weight = 2
    properties = ""
    ability = 'strength'


class Warhammer(Weapon):
    name = "Warhammer"
    cost = "15 gp"
    base_damage = "1d8"
    damage_type = "bludgeoning"
    weight = 2
    properties = "Versatile (1d10)"
    ability = 'strength'


class Whip(Weapon):
    name = "Whip"
    cost = "2 gp"
    base_damage = "1d4"
    damage_type = "slashing"
    weight = 3
    properties = "Finesse, reach"
    is_finesse = True
    ability = 'strength'


class Blowgun(Weapon):
    name = "Blowgun"
    cost = "10 gp"
    base_damage = "1"
    damage_type = "piercing"
    weight = 1
    properties = "Ammunition (range 25/100), loading"
    ability = 'dexterity'


class HandCrossbow(Weapon):
    name = "Crossbow, hand"
    cost = "75 gp"
    base_damage = "1d6"
    damage_type = "piercing"
    weight = 3
    properties = "Ammunition (range 30/120), light, loading"
    ability = 'dexterity'


class HeavyCrossbow(Weapon):
    name = "Crossbow, heavy"
    cost = "50 gp"
    base_damage = "1d10"
    damage_type = "piercing"
    weight = 18
    properties = "Ammunition (range 100/400), heaving, loading, two-handed"
    ability = 'strength'


class Longbow(Weapon):
    name = "Longbow"
    cost = "50 gp"
    base_damage = "1d8"
    damage_type = "piercing"
    weight = 2
    properties = "Ammunition (range 150/600), heavy, two-handed"
    ability = 'strength'


class Net(Weapon):
    name = "Net"
    cost = "1 gp"
    base_damage = "-"
    damage_type = ""
    weight = 3
    properties = "Special, thrown (range 5/15)"
    ability = 'strength'


# Some lists of weapons for easy proficiency resolution
simple_melee_weapons = (Club, Dagger, Greatclub, Handaxe, Javelin,
                        LightHammer, Mace, Quarterstaff, Sickle, Spear)
simple_ranged_weapons = (LightCrossbow, Dart, Shortbow, Sling)
simple_weapons = simple_melee_weapons + simple_ranged_weapons

martial_melee_weapons = (Battleaxe, Flail, Glaive, Greataxe,
                         Greatsword, Halberd, Lance, Longsword, Maul,
                         Morningstar, Pike, Rapier, Scimitar,
                         Shortsword, ThrowingHammer, Trident, WarPick,
                         Warhammer, Whip)
martial_ranged_weapons = (Blowgun, HandCrossbow, HeavyCrossbow,
                          Longbow, Net)
martial_weapons = martial_melee_weapons + martial_ranged_weapons



#!/usr/bin/env python

"""Launch a system to interactively create a character."""

import logging
# logging.basicConfig(filename='character_creater.log', level=logging.DEBUG)
log = logging.getLogger(__name__)

import math
import os
from random import randint
import subprocess

import npyscreen
import jinja2

from dungeonsheets import character, race, dice, background

char_classes = {
    'Barbarian': character.Barbarian,
    'Bard': character.Bard,
    'Cleric': character.Cleric,
    'Druid': character.Druid,
    'Fighter': character.Fighter,
    'Monk': character.Monk,
    'Paladin': character.Paladin,
    'Ranger': character.Ranger,
    'Rogue': character.Rogue,
    'Sorceror': character.Sorceror,
    'Warlock': character.Warlock,
    'Wizard': character.Wizard
}

races = {
    'Hill Dwarf': race.HillDwarf,
    'Mountain Dwarf': race.MountainDwarf,
    'High Elf': race.HighElf,
    'Wood Elf': race.WoodElf,
    'Dark Elf': race.DarkElf,
    'Lightfoot Halfling': race.LightfootHalfling,
    'Stout Halfling': race.StoutHalfling,
    'Human': race.Human,
    'Dragonborn': race.Dragonborn,
    'Gnome': race.Gnome,
    'Forest Gnome': race.ForestGnome,
    'Rock Gnome': race.RockGnome,
    'Half-Elf': race.HalfElf,
    'Half-Orc': race.HalfOrc,
    'Tiefling': race.Tiefling,
}


backgrounds = (background.Acolyte, background.Charlatan,
               background.Criminal, background.Spy,
               background.Entertainer, background.Gladiator,
               background.FolkHero, background.GuildArtisan,
               background.GuildMerchant, background.Hermit,
               background.Noble, background.Knight,
               background.Outlander, background.Sage,
               background.Sailor, background.Pirate,
               background.Soldier, background.Urchin)
backgrounds = {bg.name: bg for bg in backgrounds}


class App(npyscreen.NPSAppManaged):
    # STARTING_FORM = 'SKILLS'
    character = None
    
    def save_character(self):
        # Create the template context
        context = dict(
            char=self.character
        )
        # Render the template
        src_path = os.path.dirname(__file__)
        src_filename = 'character_template.txt'
        text = jinja2.Environment(
            loader=jinja2.FileSystemLoader(src_path or './')
        ).get_template(src_filename).render(context)
        # Save the file
        filename = self.getForm("SAVE").filename.value
        with open(filename, mode='w') as f:
            f.write(text)
        # Create the PDF character sheet
        if self.getForm('SAVE').make_pdf.value:
            log.debug("Creating PDF")
            subprocess.call(['makesheets', filename])
    
    @property
    def character_class(self, *args, **kwargs):
        return self.character_class
    
    @character_class.setter
    def character_class(self, NewClass):
        log.debug("Replacing character")
        basic_info = self.getForm('MAIN')
        self.character = NewClass(
            name=basic_info.name.value,
            player_name=basic_info.player_name.value,
            level=int(basic_info.level.value),
            strength=-1, dexterity=-1, constitution=-1,
            intelligence=-1, wisdom=-1, charisma=-1)
        self.update_max_hp()
        # Reset form widgets
        log.debug("Resetting forms")
        self.getForm('ABILITIES').reset()
    
    def update_max_hp(self):
        # Update max HP based on the class
        max_hp_fld = self.getForm('ABILITIES').max_hp
        if max_hp_fld.value == '':
            # Calculate the new value
            hit_dice = dice.read_dice_str(self.character.hit_dice)
            const = self.character.constitution.modifier
            max_hp = hit_dice.faces + const
            for d in range(hit_dice.num - 1):
                max_hp += math.ceil(hit_dice.faces/2) + const
            log.debug("Updating max hp: %d", max_hp)
            max_hp_fld.value = str(max_hp)
    
    def onStart(self):
        self.character = character.Character()
        self.addForm("MAIN", BasicInfoForm, name="Basic Info:")
        self.addForm("CLASS", CharacterClassForm, name="Select your character's class:")
        self.addForm("RACE", RaceForm, name="Select your character's race:")
        self.addForm("ALIGNMENT", AlignmentForm, name="Select your character's alignment:")
        self.addForm("ABILITIES", AbilityScoreForm, name="Choose ability scores:")
        self.addForm("BACKGROUND", BackgroundForm, name="Choose background:")
        self.addForm("SKILLS", SkillForm, name="Choose skill proficiencies")
        self.addForm("SAVE", SaveForm, name="Save character:")


class SkillForm(npyscreen.ActionForm):
    def while_editing(self):
        # Update the static skills for race and background
        bg_skills = self.parentApp.character.background.skill_proficiencies
        self.bg_skills.value = str(bg_skills)[1:-1].replace("'", "")
        race_skills = self.parentApp.character.race.skill_proficiencies
        self.race_skills.value = str(race_skills)[1:-1].replace("'", "")
        # Now set the available discretionary choices
        choices = self.parentApp.character.class_skill_choices
        static_skills = bg_skills + race_skills
        choices = (c for c in choices if c.lower() not in static_skills)
        self.skill_proficiencies.set_values(tuple(choices))
        self.update_remaining()
    
    def update_remaining(self, widget=None):
        num_choices = self.parentApp.character.num_skill_choices
        num_selected = len(self.skill_proficiencies.value)
        remaining =  num_choices - num_selected
        log.debug(f'Remaining: {remaining}')
        self.remaining.value = str(remaining)
        self.display()
    
    def create(self):
        self.bg_skills = self.add(
            npyscreen.TitleText, name="Background:",
            value="", editable=False)
        self.race_skills = self.add(
            npyscreen.TitleText, name="Racial:",
            value="", editable=False)
        self.remaining = self.add(
            npyscreen.TitleText, name="Remaining:",
            value=0, editable=False)
        self.skill_proficiencies = self.add(
            npyscreen.TitleMultiSelect, name="Skill Proficiencies:",
            values=self.parentApp.character.class_skill_choices,
            value_changed_callback=self.update_remaining)
    
    def on_ok(self):
        new_skills = self.skill_proficiencies.get_selected_objects()
        if new_skills is not None:
            new_skills = tuple(s.lower() for s in new_skills)
        else:
            new_skills = ()
        bg_skills = tuple(self.parentApp.character.background.skill_proficiencies)
        race_skills = tuple(self.parentApp.character.race.skill_proficiencies)
        all_skills = new_skills + bg_skills + race_skills
        self.parentApp.character.skill_proficiencies = all_skills
        log.debug(f"Skill proficiencies: {all_skills}")
        self.parentApp.setNextForm('SAVE')
    
    def on_cancel(self):
        self.parentApp.setNextForm('BACKGROUND')


class AbilityScoreForm(npyscreen.ActionForm):
    def roll_dice(self):
        """Get six ability scores that can then be assigned to abilities.""" 
        def roll_score():
            # Roll 4 dice and add the 3 highest
            rolls = (randint(1, 6) for i in range(4))
            score = sum(sorted(rolls)[-3:])
            return score
        scores = (roll_score() for i in range(6))
        return tuple(sorted(scores, reverse=True))
    
    def reset(self):
        # Update the character in real time
        attrs = ('strength', 'dexterity', 'constitution',
                 'intelligence', 'wisdom', 'charisma')
        for attr in attrs:
            getattr(self, attr).value = ''
    
    def while_editing(self):
        # Update the character in real time
        attrs = ('strength', 'dexterity', 'constitution',
                 'intelligence', 'wisdom', 'charisma')
        for attr in attrs:
            fld = getattr(self, attr)
            try:
                race_bonus = getattr(self.parentApp.character.race, f'{attr}_bonus')
                val = int(float(fld.value))
            except ValueError:
                # Not an integer, so clear the field
                fld.value = ''
            else:
                # Valid number, so process it
                curr_val = getattr(self.parentApp.character, attr).value
                if val != curr_val:
                    log.debug("Setting %s to %s", attr, str(val))
                    setattr(self.parentApp.character, attr, val)
                    # Update the "character" with new values
                    if attr == 'constitution':
                        self.parentApp.update_max_hp()
                        fld.value = str(val)
        # Update the form display
        self.display()
        
    def create(self):
        self.score_options = self.add(
            npyscreen.TitleFixedText, name="Rolls:", editable=False,
            value=str(self.roll_dice())[1:-1])
        self.add(npyscreen.FixedText, editable=False,
                 value="Take the six rolls and assign each one to an ability.")
        self.add(npyscreen.FixedText, editable=False,
                 value="Do not add racial bonuses, they will be added for you.")
        self.strength = self.add(npyscreen.TitleText, name="Strength:")
        self.dexterity = self.add(npyscreen.TitleText, name="Dexterity:")
        self.constitution = self.add(npyscreen.TitleText, name="Constitution:")
        self.intelligence = self.add(npyscreen.TitleText, name="Intelligence:")
        self.wisdom = self.add(npyscreen.TitleText, name="Wisdom:")
        self.charisma = self.add(npyscreen.TitleText, name="Charisma:")
        self.add(npyscreen.FixedText, editable=False,
                 value="Maximum hit points initially determined by constitution.")
        self.max_hp = self.add(npyscreen.TitleText, name="Max HP:")
    
    def on_ok(self):
        self.parentApp.setNextForm('BACKGROUND')
    
    def on_cancel(self):
        self.parentApp.setNextForm('ALIGNMENT')


class CharacterClassForm(npyscreen.ActionForm):
    def create(self):
        self.character_class = self.add(
            npyscreen.TitleMultiLine, name="Class:", values=tuple(char_classes.keys()))
    
    def on_ok(self):
        if self.character_class.value is not None:
            selected_class = self.character_class.values[self.character_class.value]
            selected_class = char_classes[selected_class]
            log.debug('Selected character class %s', selected_class.class_name)
            self.parentApp.character_class = selected_class
            self.parentApp.setNextForm('RACE')
    
    def on_cancel(self):
        self.parentApp.setNextForm('MAIN')


class BackgroundForm(npyscreen.ActionForm):

    def create(self):
        self.background = self.add(
            npyscreen.TitleMultiLine,
            name="Background:", values=tuple(backgrounds.keys()))
    
    def on_ok(self):
        if self.background.value is not None:
            selected_bg = self.background.values[self.background.value]
            Background = backgrounds[selected_bg]
            self.parentApp.character.background = Background()
            # Update the languages based on background and race
            race_languages = self.parentApp.character.race.languages
            languages = Background.languages + race_languages
            self.parentApp.character.languages = ', '.join(languages)
            log.debug("Selected character background: %s", Background.name)
            self.parentApp.setNextForm('SKILLS')
    
    def on_cancel(self):
        self.parentApp.setNextForm('ABILITIES')


class RaceForm(npyscreen.ActionForm):
    def create(self):
        self.race = self.add(
            npyscreen.TitleMultiLine, name="Race:", values=tuple(races.keys()))
    
    def on_ok(self):
        if self.race.value is not None:
            selected_race = self.race.values[self.race.value]
            SelectedRace = races[selected_race]
            log.debug('Selected character race: %s', SelectedRace.name)
            self.parentApp.character.race = SelectedRace()
            self.parentApp.setNextForm('ALIGNMENT')
    
    def on_cancel(self):
        self.parentApp.setNextForm('CLASS')
        


class AlignmentForm(npyscreen.ActionForm):
    """Choose your character's alignment."""
    alignments = ('Lawful good', 'Neutral good', 'Chaotic good',
                  'Lawful neutral', 'True neutral', 'Chaotic neutral',
                  'Lawful evil', 'Neutral evil', 'Chaotic evil', )
    
    def create(self):
        self.alignment = self.add(
            npyscreen.TitleMultiLine, name="Alignment:", values=self.alignments)
    
    def on_ok(self):
        if self.alignment.value is not None:
            selected_alignment = self.alignment.values[self.alignment.value]
            log.debug('Selected character alignment %s', selected_alignment)
            self.parentApp.character.alignment = selected_alignment
            self.parentApp.setNextForm('ABILITIES')
    
    def on_cancel(self):
        self.parentApp.setNextForm('RACE')


class BasicInfoForm(npyscreen.ActionForm):
    def create(self):
        self.name = self.add(
            npyscreen.TitleText, name="Character Name:", use_two_lines=False)
        self.player_name = self.add(
            npyscreen.TitleText, name="Player Name:", use_two_lines=False)
        self.level = self.add(
            npyscreen.TitleText, name='Level:', value="1")
    
    def on_ok(self):
        # Update the default filename
        name = self.name.value
        if name == '':
            filename = 'new_character.py'
        else:
            filename = f'{name.split(" ")[0].lower()}.py'
        save_form = self.parentApp.getForm('SAVE')
        save_form.filename.value = filename
        # Move to the next form
        self.parentApp.setNextForm('CLASS')
    
    def on_cancel(self):
        self.parentApp.setNextForm(None)


class SaveForm(npyscreen.ActionForm):
    def create(self):
        self.filename = self.add(
            npyscreen.TitleText, name='Filename:')
        self.make_pdf = self.add(npyscreen.Checkbox, name="Create PDF:", value=True)
        self.instructions = self.add(
            npyscreen.FixedText, editbale=False,
            value="After saving, edit this file to finish your personality, etc.")
    
    def on_ok(self):
        self.parentApp.setNextForm(None)
    
    def on_cancel(self):
        self.parentApp.setNextForm('SKILLS')


def main():
    my_app = App()
    
    try:
        my_app.run()
    except KeyboardInterrupt:
        log.error("Aborted by user request")
    else:
        my_app.save_character()


if __name__ == '__main__':
    main()



from . import weapons


__all__ = ('Dwarf', 'HillDwarf', 'MountainDwarf', 'Elf', 'HighElf',
           'WoodElf', 'DarkElf', 'Halfling', 'LightfootHalfling',
           'StoutHalfling', 'Human', 'Dragonborn', 'Gnome', 'ForestGnome',
           'RockGnome', 'HalfElf', 'HalfOrc', 'Tiefling')


class Race():
    name = "Unknown"
    size = "medium"
    speed = 30
    languages = ('Common', )
    proficiencies_text = tuple()
    weapon_proficiences = tuple()
    skill_proficiencies = ()
    strength_bonus = 0
    dexterity_bonus = 0
    constitution_bonus = 0
    intelligence_bonus = 0
    wisdom_bonus = 0
    charisma_bonus = 0
    hit_point_bonus = 0
    
    def __str__(self):
        return self.name
    
    def __repr__(self):
        return f"<self.name>"


# Dwarves
class Dwarf(Race):
    name = "Dwarf"
    size = "medium"
    speed = 25
    languages = ("Common", "Dwarvish")
    constitution_bonus = 2
    proficiencies_text = ('battleaxes', 'handaxes', 'throwing hammers', 'warhammers')
    weapon_proficiences = (weapons.Battleaxe, weapons.Handaxe,
                           weapons.ThrowingHammer, weapons.Warhammer)


class HillDwarf(Dwarf):
    name = "Hill Dwarf"
    wisdom_bonus = 1
    hit_point_bonus = 1


class MountainDwarf(Dwarf):
    name = "Mountain Dwarf"
    strength_bonus = 2


# Elves
class Elf(Race):
    name = "Elf"
    size = "medium"
    speed = 30
    dexterity_bonus = 2
    skill_proficiencies = ('perception',)
    languages = ('Common', 'Elvish')


class HighElf(Elf):
    name = "High Elf"
    weapon_proficiencies = (weapons.Longsword, weapons.Shortsword,
                           weapons.Shortbow, weapons.Longbow)
    proficiencies_text = ('longswords', 'shortswords', 'shortbows', 'longbows')
    intelligence_bonus = 1
    languages = ('Common', 'Elvish', '[choose one]')


class WoodElf(Elf):
    name = "Wood Elf"
    weapon_proficiencies = (weapons.Longsword, weapons.Shortsword,
                           weapons.Shortbow, weapons.Longbow)
    proficiencies_text = ('longswords', 'shortswords', 'shortbows', 'longbows')
    wisdom_bonus = 1


class DarkElf(Elf):
    name = "Dark Elf"
    weapon_proficiencies = (weapons.Rapier, weapons.Shortsword, weapons.HandCrossbow)
    proficiencies_text = ('rapiers', 'shortswords', 'hand crossbows')
    charisma_bonus = 1


# Halflings
class Halfling(Race):
    name = "Halfling"
    size = "small"
    speed = 25
    dexterity_bonus = 2
    languages = ('Common', 'Halfling')


class LightfootHalfling(Halfling):
    name = "Lightfoot Halfling"
    charisma_bonus = 1


class StoutHalfling(Halfling):
    name = "Stout Halfling"
    constitution_bonus = 1


# Humans
class Human(Race):
    name = "Human"
    size = "medium"
    speed = 30
    strength_bonus = 1
    dexterity_bonus = 1
    constitution_bonus = 1
    intelligence_bonus = 1
    wisdom_bonus = 1
    charisma_bonus = 1
    languages = ("Common", '[choose one]')


# Dragonborn
class Dragonborn(Race):
    name = "Dragonborn"
    size = "medium"
    speed = 30
    strength_bonus = 2
    charisma_bonus = 1
    languages = ("Common", "Draconic")


# Gnomes
class Gnome(Race):
    name = "Gnome"
    size = "small"
    speed = 25
    intelligence_bonus = 2
    languages = ("Common", "Gnomish")


class ForestGnome(Gnome):
    name = "Forest Gnome"
    dexterity_bonus = 1


class RockGnome(Gnome):
    name = "Rock Gnome"
    constitution_bonus = 1


# Half-elves
class HalfElf(Race):
    name = "Half-Elf"
    size = "medium"
    speed = 30
    charisma_bonus = 2
    languages = ("Common", "Elvish", "[choose one]")


# Half-Orcs
class HalfOrc(Race):
    name = "Half-Orc"
    size = "medium"
    speed = 30
    strength_bonus = 2
    constitution_bonus = 1
    languages = ("Common", "Orc")


# Tielflings
class Tiefling(Race):
    name = "Tiefling"
    size = "medium"
    speed = 30
    intelligence_bonus = 1
    charisma_bonus = 2
    languages = ("Common", "Infernal")
